#===============================================================================
# Program:      subs_cacti.pl
#
# Description:  Library of miscellaneous subroutines created for cacti.
#
#               This script was developed under ActiveState Perl version 5.6.1.
#
# Requirements: This ActiveState PERL must be installed on the system where the
#               script is executed.
#
#               The following additional Perl modules must also be installed:
#                   None
#
#               The following files must exist in the same directory as this
#               script:
#                   None
#
#               Known limitations:
#                   None
#
#
# Date       Author/Email                Description
# ---------- --------------------------- ---------------------------------------
# 03/17/2003 Jim Lancaster               Initial version 0.1
#            jlancaster@mbdworld.com
#
# 03/17/2003 Jim Lancaster               Version 0.1a
#                                        Minor change to createDS,graphIf
#
#
# Known issues:
#    1. None
#
# Future plans:
#    1. None
#
#===============================================================================
use strict;
return 1;

#===============================================================================
#                       Global Variables
#===============================================================================
my ($newSrcID,%newFieldID);             # Set by createDIS()
my ($ds,$newDSID);                      # Set by createDS()
my (%rrdGraphID);                       # Set by createGraph()

#===============================================================================
#                       Subroutines
#===============================================================================
# cactiDB()
#
# Description - Read cacti database settings from config.php.
#
# Inputs:
#  1.
#
# Processing:
#  1. Open config.php, grep the MySQL database settings and store in a hash.
#
# Outputs:
#  1. %cactiDB - A hash containing the cacti MySQL database settings.
#
#===============================================================================
sub cactiDB
{

    my %cactiDB;

    # The relative location of 'config.php' depends upon where
    #  the script was called.  It should be in either of these
    #  two places.
    open(CFGFILE,"<../include/config.php")
        or open(CFGFILE,"<include/config.php")
            or die "\n\nUnable to open configuration file!";

    my @dblines = grep{/^\$database/} readline(<CFGFILE>);
    foreach my $line (@dblines) {
        chop($line);
        chop($line);
        my ($key,$value) = split / = /,$line;
        my ($hash_name,$new_key) = split /_/,$key;
        $value =~ s/\"//g;
        $cactiDB{$new_key} = $value;
    }

    # For testing purposes only.
    #&dref(\%cactiDB,"hash");

    return (%cactiDB);
}

#===============================================================================
# createDIS($dbh,$disName,$ipaddress,$script,$outputStr)
#
# Description - Create the Data Input Source (src & src_fields) records in cacti
#
# Inputs:
#  1.$dbh           - Database handle to cacti database.
#  2.$disName       - The DIS name.
#  3.$ipaddress     - The IP address (or dns hostname) of the device.
#  4.$scriptname    - The name of the perl script that will be used to poll this device.
#  4.$outputStr     - A reference to an array containing the output field descriptions.
#
# Processing:
#  1.
#
# Outputs:
#  1.$newSrcID      - The new src record ID
#  2.\%newFieldID   - A reference to a hash containing the new src_fields record IDs.
#
#===============================================================================
sub createDIS
{

    my ($dbh,$disName,$ipaddress,$script,$outputStr) = @_;

    my ($query,$sth,$sth2);
    my ($newSrcID,%inFieldID,%outFieldID);

    # Test to see if this DIS already exists. Reuse it if it does.
    $query = "SELECT ID FROM src WHERE Name = '$disName'";
    my $result = $dbh->selectrow_array ($query);
    if (defined($result)) {

        # Save the src record ID.
        $newSrcID = $result;

        # Query for and save the corresponding src_fields ID records
        $query = "SELECT ID,Name FROM src_fields WHERE SrcID = $newSrcID and InputOutput = ?";
        $sth = $dbh->prepare ($query);
        $sth->execute ('in')
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
        while (my $ref = $sth->fetchrow_arrayref) {
            $inFieldID{$$ref[1]} = $$ref[0];
        }
        $sth->execute ('out')
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
        while (my $ref = $sth->fetchrow_arrayref) {
            $outFieldID{$$ref[1]} = $$ref[0];
        }

        # Close the statement handle.
        $sth->finish();

    } else {

        # If it doesn't already exist, start by creating src table record.
        my $str = "<".join("> <",@$outputStr).">";

        $query = "INSERT INTO src (Name,FormatStrIn,FormatStrOut)
            VALUES ('$disName','perl <path_cacti>/scripts/$script','$str')";
        $dbh->do ($query)
            or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";

        # Save the new src record ID.
        $newSrcID = $dbh->{mysql_insertid};

        # Next, create the src_fields table records
        $query = "INSERT INTO src_fields (SrcID,Name,DataName,InputOutput,UpdateRRA)
            VALUES (?,?,?,?,?)";
        $sth = $dbh->prepare ($query)
            or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";

        # Search through the $script for input fields
        $_ = $script;
        my @inputs = m/<\w+>/g;

        # Create the 'input' records.
        foreach my $value (@inputs) {

            # Strip off the "<>"
            $value =~ s/[<|>]//g;

            $sth->execute ($newSrcID,$value,$value,'in','')
                or die "\n\nUnable to execute statement handle\n$DBI::errstr";

            # Save the new src_fields record ID.
            $inFieldID{$value} = $dbh->{mysql_insertid};
        }

        # Create the 'output' records from @outputStr.
        foreach my $value (@$outputStr) {

            $sth->execute ($newSrcID,$value,$value,'out','on')
                or die "\n\nUnable to execute statement handle\n$DBI::errstr";

            # Save the new src_fields record ID.
            $outFieldID{$value} = $dbh->{mysql_insertid};
        }

        # Close the statement handle.
        $sth->finish();
    }

    # For testing purposes only
    #print "$newSrcID\n";
    #&dref(\%inFieldID,'hash');
    #&dref(\%outFieldID,'hash');

    return ($newSrcID,\%inFieldID,\%outFieldID);

}

#===============================================================================
# createDS($dbh,$hostname,$newSrcID,$inFieldID,$outFieldID,$srcInput,$srcData)
#
# Description - Create the Data Source (rrd_ds,lnk_ds_rra & src_data) records
#
# Inputs:
#  1.$dbh           - Database handle to cacti database.
#  3.$hostname      - The name of the device. This will be the name under which the device
#                      will be created. The default is the sysName for SNMP devices and
#                      the Windows machine name for WMI-polled devices.
#  4.$newSrcID      - The new src record ID.
#  5.$inFieldID     - A reference to a hash containing the new input src_fields record IDs..
#  6.$outFieldID    - A reference to a hash containing the new output src_fields record IDs..
#  7.$srcInput      - An array of references to hashes of hashes of performance data
#  6.$srcData       - An array of references to hashes of hashes of performance data
#                      The elements are assigned as follows:
#                       $$srcData[0] = cpu
#                       $$srcData[1] = memory
#                       $$srcData[2] = disk
#                       $$srcData[3] = interface
#                       $$srcData[4] = printer
#                       $$srcData[5] = rmon
#
# Processing:
#  1.
#
# Outputs:
#  1.\%ds           - A reference to a hash containing rrd_ds record IDs.
#  2.\%newDSID      - A reference to a hash containing...
#
#=============================================================================
sub createDS
{

    my ($dbh,$hostname,$newSrcID,$inFieldID,$outFieldID,$srcInput,$srcData) = @_;

    my ($query,$sth,$sth2);

    # Create a hash of default rrd_ds field values for the 'parent'
    #  record.  We will modify these slightly for subsequent 'child' records.
    my %ds;
    $ds{'SubDSID'}          = 0;                                # n/a on DS parent.
    $ds{'SubFieldID'}       = 0;                                # n/a on DS parent.
    $ds{'Name'}             = uc($hostname);                    #
    $ds{'DataSourceTypeID'} = 1;                                # Gauge
    $ds{'Heartbeat'}        = 600;                              # Default
    $ds{'MinValue'}         = 0;                                # Default
    $ds{'MaxValue'}         = 100000000000;                     # A very large number.
    $ds{'SrcID'}            = $newSrcID;                        # The newly created DIS.
    $ds{'Active'}           = "on";                             # Default
    $ds{'DSName'}           = lc($ds{'Name'});                  #
    $ds{'DSPath'}           = "<path_rra>/".$ds{'DSName'}.".rrd"; #
    $ds{'Step'}             = 300;                              # Default.
    $ds{'IsParent'}         = 1;                                # Parent record flag.

    $query = "INSERT INTO rrd_ds (SubDSID,SubFieldID,Name,DataSourceTypeID,Heartbeat,
        MinValue,MaxValue,SrcID,Active,DSName,DSPath,Step,IsParent)
        VALUES ($ds{'SubDSID'},$ds{'SubFieldID'},'$ds{'Name'}',$ds{'DataSourceTypeID'},$ds{'Heartbeat'},
        $ds{'MinValue'},$ds{'MaxValue'},$ds{'SrcID'},'$ds{'Active'}','$ds{'DSName'}','$ds{'DSPath'}',
        $ds{'Step'},$ds{'IsParent'})";
    $dbh->do ($query)
        or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";

    # Save the new rrd_ds record id
    my $parentID = $dbh->{mysql_insertid};

    # Link the parent record to the 4 standard rra's.
    $query = "INSERT INTO lnk_ds_rra (DSID,RRAID) VALUES ($parentID,?)";
    $sth = $dbh->prepare ($query)
                or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";
    for (my $i = 1;$i < 5;$i++) {
        $sth->execute ($i)
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
    }

    # Close the statement handle.
    $sth->finish();


    # Modify the default rrd_ds field values for subsequent 'child' records.
    my %newDSID;
    $ds{'SubDSID'}          = $parentID;            #
    #$ds{'Name'}                                    # Different for each record.
    $ds{'SrcID'}            = 0;                    # n/a for child record.
    #$ds{'DSName'}                                  # Different for each record.
    $ds{'Step'}             = 0;                    # n/a for child record.
    $ds{'IsParent'}         = 0;                    # n/a for child record.

    # Create the rrd_ds 'child' records.
    $query = "INSERT INTO rrd_ds (SubDSID,SubFieldID,Name,DataSourceTypeID,Heartbeat,
        MinValue,MaxValue,SrcID,Active,DSName,DSPath,Step,IsParent)
        VALUES ($ds{'SubDSID'},?,?,?,$ds{'Heartbeat'},
        $ds{'MinValue'},?,$ds{'SrcID'},'$ds{'Active'}',?,'$ds{'DSPath'}',
        $ds{'Step'},$ds{'IsParent'})";
    $sth = $dbh->prepare ($query)
        or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";

    foreach my $key (sort keys %$outFieldID) {

        # Set the internal DS names used by both Cacti and rrd.  The cacti DS name
        #  needs to be prefixed by the host name for Cacti.  The rrd internal DS
        #  name cannot be longer than 19 characters. (Thanks bulek!)
        $ds{'DSName'}   = substr($key,0,19);
        $ds{'Name'}     = uc($hostname)."-".$ds{'DSName'};

        # Modify the default values for DS type and max value.
        if ($key =~ /^BytesFreeDisk/) {
            $ds{'DataSourceTypeID'} = 1;            # Gauge
            my ($temp,$outputID)    = split("-",$key);
            $ds{'MaxValue'}         = ${@$srcData[2]}{$outputID}{'Size'};
        } elsif ($key =~ /^BytesFreeMem/) {
            $ds{'DataSourceTypeID'} = 1;            # Gauge
            $ds{'MaxValue'}         = ${@$srcData[1]}{'Memory'}{'TotalPhysicalMemory'};
        } elsif ($key =~ /^BytesFreePageFile/) {
            $ds{'DataSourceTypeID'} = 1;            # Gauge
            $ds{'MaxValue'}         = ${@$srcData[1]}{'Memory'}{'TotalPageFileSpace'};
        } elsif ($key =~ /^BytesFreeVirtMem/) {
            $ds{'DataSourceTypeID'} = 1;            # Gauge
            $ds{'MaxValue'}         = ${@$srcData[1]}{'Memory'}{'TotalVirtualMemory'};
        } elsif ($key =~ /^Percent/) {
            $ds{'DataSourceTypeID'} = 1;            # Gauge
            $ds{'MaxValue'}         = 100;          # 100%
        } elsif ($key =~ /^[if|rmon]/) {
            $ds{'DataSourceTypeID'} = 2;            # Counter
            $ds{'MaxValue'}         = 100000000000; # A very large number.
        } elsif ($key =~ /^prtMarkerLifeCount/) {
            $ds{'DataSourceTypeID'} = 2;            # Counter
            $ds{'MaxValue'}         = 100000000000; # A very large number.
        } else {
            die ("Unknown Data Source Type for $key! \n");
        }

        $sth->execute ($$outFieldID{$key},$ds{'Name'},$ds{'DataSourceTypeID'},$ds{'MaxValue'},$ds{'DSName'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";

        # Save the new rrd_ds record ID.
        $newDSID{$key} = $dbh->{mysql_insertid};

    }

    # Close the statement handle.
    $sth->finish();

    # Finally, create the src_data records.
    $query = "INSERT INTO src_data (FieldID,DSID,Value) VALUES (?,?,?)";
    $sth = $dbh->prepare ($query)
        or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";

    # Start with the input fields
    foreach my $key (keys %$srcInput) {

        $sth->execute ($$inFieldID{$key},$parentID,$$srcInput{$key})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
    }

    # Cycle through the source data array and create the src_data records for the output fields.
    foreach my $ref (@$srcData) {
        foreach my $key1 (sort keys %$ref) {

            # Creating the src_data record from the reference to the "single-value" hashes of
            #  hashes is straightforward. Key2 is common to the hash key, outFieldID, newDSID.
            if ($key1 eq "Memory") {
                foreach my $key2 (sort keys %{$$ref{$key1}}) {
                    if ($key2 =~ /^(Bytes|Percent)/) {
                        $sth->execute ($$outFieldID{$key2},$newDSID{$key2},$$ref{$key1}{key2})
                            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
                    }
                }

            } elsif ($key1 eq "Printer") {
                foreach my $key2 (sort keys %{$$ref{$key1}}) {
                    if ($key2 =~ /^prtMarkerLifeCount/) {
                        $sth->execute ($$outFieldID{$key2},$newDSID{$key2},$$ref{$key1}{key2})
                            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
                    }
                }

            } elsif ($key1 eq "if") {
                foreach my $key2 (sort keys %{$$ref{$key1}}) {
                    if ($key2 !~ /^(ifDescr|ifIndex|ifOperStatus|ifType)/) {
                        $sth->execute ($$outFieldID{$key2},$newDSID{$key2},$$ref{$key1}{key2})
                            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
                    }
                }

            } elsif ($key1 eq "rmon") {
                foreach my $key2 (sort keys %{$$ref{$key1}}) {
                    if ($key2 !~ /^(rmonDataSource|rmonIndex)/) {
                        $sth->execute ($$outFieldID{$key2},$newDSID{$key2},$$ref{$key1}{key2})
                            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
                    }
                }

            # For servers it is a little more complicated because there can be more than one
            #  cpu, disk & interface . When we created the newField and newDSID key/value pairs,
            #  we essentially converted the 3-dimensional cpu, disk & interface hashes of
            #  hashes into a 2-dimensional hashes by combining key2 and key1.
            #  For example, "PercentCPULoad-0" is the PercentCPULoad ($key2) for CPU #0 ($key1).
            #  If there is more than one CPU, the combined keys would be "PercentCPULoad-1",
            #  "PercentCPULoad-2", etc. Now, to create the src_data records we have to do it
            #  again to link the 2-dimensional outFieldID & newDSID hashes to the 3-dimensional
            #  cpu, disk & interface hashes of hashes.
            } else {
                foreach my $key2 (sort keys %{$$ref{$key1}}) {
                    if ($key2 =~ /^(Bytes|Percent|ifIn|ifOut)/) {
                        $sth->execute ($$outFieldID{$key2."-".$key1},$newDSID{$key2."-".$key1},$$ref{$key1}{$key2})
                            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
                    }
                }
            }
        }
    }

    # Close the statement handle.
    $sth->finish();

    # For testing purposes only
    #dref(\%newDSID,"hash");

    return (\%ds,\%newDSID);

}

#===============================================================================
# createGraph($dbh,$hostname,$newDSID,\@srcData)
#
# Description - Create
#
# Inputs:
#  1.$dbh       - Database handle to cacti database.
#  2.$hostname  - The name of the device. This will be the name under which the device
#                  will be created. The default is the sysName for SNMP devices and
#                  the Windows machine name for WMI-polled devices.
#  3.$newDSID   - A reference to a hash containing
#  4.$srcData   - An array of references to hashes of hashes of performance data
#                  The elements are assigned as follows:
#                   $$srcData[0] = hoh_cpu
#                   $$srcData[1] = hoh_mem
#                   $$srcData[2] = hoh_disk
#                   $$srcData[3] = hoh_if
#                   $$srcData[4] = hoh_prn
#                   $$srcData[5] = hoh_rmon
#                  Note that some elements may remain undefined.
#
# Processing:
#  1.
#
# Outputs:
#  1.\%rrdGraphID
#
#=============================================================================
sub createGraph
{

    my ($dbh,$hostname,$newDSID,$srcData) = @_;

    my ($query,$sth,$sth2);

    # Start by creating a hash of the rrd_graph default field values.
    my %graph;
    $graph{'ImageFormatID'}         = 1;            # Default (PNG format).
    #$graph{'Title'}                                # This will be defined later.
    $graph{'Height'}                = 120;          # Default
    $graph{'Width'}                 = 500;          # Default
    $graph{'UpperLimit'}            = 0;            # Default
    $graph{'LowerLimit'}            = 0;            # Default
    #$graph{'VerticalLabel'}                        # This will be defined later.
    $graph{'AutoScale'}             = "";           # "on" or "" (Off)
    $graph{'AutoPadding'}           = "on";         # Default
    $graph{'AutoScaleOpts'}         = 2;            # Default (--alt-autoscale-max)
    $graph{'Rigid'}                 = "";           # Off; set to "on" to turn on.
    $graph{'BaseValue'}             = 1000;         # Default.
    $graph{'Grouping'}              = "";           # Default.
    $graph{'Export'}                = "on";         # Default.
    #$graph{'UnitValue'}                            # not used.
    $graph{'UnitExponentValue'}     = 0;            # Default.
    $graph{'AutoScaleLog'}          = "on";         # "on" or "" (Off)

    # Create a statement handle for the rrd_graph records.
    $query = "INSERT INTO rrd_graph (ImageFormatID,Title,Height,Width,UpperLimit,
        LowerLimit,VerticalLabel,AutoScale,AutoPadding,AutoScaleOpts,Rigid,BaseValue,
        Grouping,Export,UnitExponentValue,AutoScaleLog)
        VALUES ($graph{'ImageFormatID'},?,$graph{'Height'},$graph{'Width'},
        $graph{'UpperLimit'},$graph{'LowerLimit'},?,'$graph{'AutoScale'}',
        '$graph{'AutoPadding'}',$graph{'AutoScaleOpts'},'$graph{'Rigid'}',
        $graph{'BaseValue'},'$graph{'Grouping'}','$graph{'Export'}',
        $graph{'UnitExponentValue'},'$graph{'AutoScaleLog'}')";
    $sth = $dbh->prepare ($query)
        or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";

    # Create a statement handle for the rrd_graph_item records.
    $query = "INSERT INTO rrd_graph_item (DSID,ColorID,TextFormat,Sequence,GraphID,
        GraphTypeID,ConsolidationFunction,HardReturn,CDEFID,SequenceParent,Parent,GprintOpts)
        VALUES (?,?,?,?,?,?,?,?,?,?,?,?)";
    $sth2 = $dbh->prepare ($query)
        or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";

    # Set the defaults for all rrd_graph_item records.
    my %graphItem;
    $graphItem{'CDEFID'}            = 0;            # Default.
    $graphItem{'SequenceParent'}    = 0;            # Default.
    $graphItem{'Parent'}            = 0;            # Default.
    $graphItem{'GprintOpts'}        = 1;            # Default

    # Create an array of references to subroutines that corresponds the the hash types.
    my @graphType;
    $graphType[0] = *graphTypeCPU{CODE};
    $graphType[1] = *graphTypeMem{CODE};
    $graphType[2] = *graphTypeDisk{CODE};
    $graphType[3] = *graphTypeIf{CODE};
    $graphType[4] = *graphTypePrn{CODE};
    $graphType[5] = *graphTypeRmon{CODE};

    my $i = 0;
    foreach my $ref (@$srcData) {
        if (values(%{$ref}) > 0) {
            &{$graphType[$i]}($dbh,$sth,$sth2,\%graph,\%graphItem,@$srcData[$i],$hostname,$newDSID);
        }
        $i++;
    }

    return (\%rrdGraphID);

}

#===============================================================================
# graphTypeCPU($dbh,$sth,$sth2,$graph,$graphItem,$hoh_if,$hostname,$newDSID)
#
# Description - Create a standard CPU graph.
#
# Inputs:
#  1.$dbh
#  2.$sth
#  3.$sth2
#  4.$graph
#  5.$graphItem
#  6.$hoh_if
#  7.$hostname
#  8.$newDSID
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#=============================================================================
sub graphTypeCPU
{

    my ($dbh,$sth,$sth2,$graph,$graphItem,$hoh_cpu,$hostname,$newDSID) = @_;

    my $count = 0;
    my $graphID;
    my $sequence = 1;

    # Create the CPU Utilization graph. There will be one for all CPUs.
    foreach my $key (sort keys %$hoh_cpu) {

        # On the first pass through, create the rrd_graph record.
        if ($count == 0) {

            # Create the rrd_graph record.
            $$graph{'Title'}                = $hostname." CPU Utilization";
            $$graph{'VerticalLabel'}        = "Percent Used";
            $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
                or die "\n\nUnable to execute statement handle\n$DBI::errstr";

            # Set the rrd_graph_item default field values.
            $$graphItem{'GraphID'}          = $dbh->{mysql_insertid};
            $$graphItem{'GraphTypeID'}      = 5; # line2

            # Save the rrd_graph record ID for use in creating the graph hierarchy.
            $rrdGraphID{'CPU'}              = $$graphItem{'GraphID'};

        }

        #########################################################################
        # Note: I need to create an array of possible graph colors so that each
        # cpu can be represented by a different color on the same graph.
        #########################################################################

        # Complete the settings for this graph item.
        $$graphItem{'DSID'}                 = $$newDSID{"PercentCPULoad-".$key};
        $$graphItem{'ColorID'}              = 21; #0000FF Blue
        $$graphItem{'TextFormat'}           = "CPU Utilization - CPU".$key;
        $$graphItem{'Sequence'}             = $sequence;

        # Create the rrd_graph_item records.
        &createGraphItems($sth2,$graphItem);

        # Increment the counter(s).
        $count++;;
        $sequence = $sequence + 4;

    }
}

#===============================================================================
# graphTypeMem($dbh,$sth,$sth2,$graph,$graphItem,$hoh_if,$hostname,$newDSID)
#
# Description - Create
#
# Inputs:
#  1.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#=============================================================================
sub graphTypeMem
{

    my ($dbh,$sth,$sth2,$graph,$graphItem,$hoh_mem,$hostname,$newDSID) = @_;

    my ($graphIDBytes,$graphIDPercent);

    # Create the BytesFree memory rrd_graph record.
    $$graph{'Title'}                = $hostname." Free Memory (Bytes)";
    $$graph{'VerticalLabel'}        = "Bytes Free";
    $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
        or die "\n\nUnable to execute statement handle\n$DBI::errstr";

    # Save the BytesFree rrd_graph record ID.
    $graphIDBytes                   = $dbh->{mysql_insertid};

    # Save the rrd_graph record ID for use in creating the graph hierarchy.
    $rrdGraphID{'MemBytes'}         = $graphIDBytes;

    # Create the PercentFree memory rrd_graph record.
    $$graph{'Title'}                = $hostname." Memory Utilization";
    $$graph{'VerticalLabel'}        = "Percent Free";
    $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
        or die "\n\nUnable to execute statement handle\n$DBI::errstr";

    # Save the PercentFree rrd_graph record ID.
    $graphIDPercent                 = $dbh->{mysql_insertid};

    # Save the rrd_graph record ID for use in creating the graph hierarchy.
    $rrdGraphID{'MemPercent'}       = $graphIDPercent;

    # Create the rrd_graph_item records.
    foreach my $key (keys %{$$hoh_mem{'Memory'}}) {

        if ($key =~ /BytesFreeMem/) {

            # Set the remaining fields for the graph item.
            $$graphItem{'DSID'}             = $$newDSID{$key};
            $$graphItem{'ColorID'}          = 10;
            $$graphItem{'TextFormat'}       = "Free Memory (Bytes)";
            $$graphItem{'Sequence'}         = 1;
            $$graphItem{'GraphID'}          = $graphIDBytes;
            $$graphItem{'GraphTypeID'}      = 7; # area

            # Create the rrd_graph_item record.
            &createGraphItems($sth2,$graphItem);

        } elsif ($key =~ /BytesFreePageFile/) {

            # Set the remaining fields for the graph item.
            $$graphItem{'DSID'}             = $$newDSID{$key};
            $$graphItem{'ColorID'}          = 7;
            $$graphItem{'TextFormat'}       = "Free Pagefile (Bytes)";
            $$graphItem{'Sequence'}         = 5;
            $$graphItem{'GraphID'}          = $graphIDBytes;
            $$graphItem{'GraphTypeID'}      = 8; # stack

            # Create the rrd_graph_item record.
            &createGraphItems($sth2,$graphItem);

        } elsif ($key =~ /BytesFreeVirtMem/) {

            # Set the remaining fields for the graph item.
            $$graphItem{'DSID'}             = $$newDSID{$key};
            $$graphItem{'ColorID'}          = 60;
            $$graphItem{'TextFormat'}       = "Free Virtual Memory (Bytes)";
            $$graphItem{'Sequence'}         = 9;
            $$graphItem{'GraphID'}          = $graphIDBytes;
            $$graphItem{'GraphTypeID'}      = 5; # line2

            # Create the rrd_graph_item record.
            &createGraphItems($sth2,$graphItem);

        } elsif ($key =~ /PercentFreeMem/) {

            # Set the remaining fields for the graph item.
            $$graphItem{'DSID'}             = $$newDSID{$key};
            $$graphItem{'ColorID'}          = 10;
            $$graphItem{'TextFormat'}       = "Memory Utilization";
            $$graphItem{'Sequence'}         = 1;
            $$graphItem{'GraphID'}          = $graphIDPercent;
            $$graphItem{'GraphTypeID'}      = 5; # line2

            # Create the rrd_graph_item record.
            &createGraphItems($sth2,$graphItem);

        } elsif ($key =~ /PercentFreePageFile/) {

            # Set the remaining fields for the graph item.
            $$graphItem{'ColorID'}          = 7;
            $$graphItem{'DSID'}             = $$newDSID{$key};
            $$graphItem{'TextFormat'}       = "Pagefile Utilization";
            $$graphItem{'Sequence'}         = 5;
            $$graphItem{'GraphID'}          = $graphIDPercent;
            $$graphItem{'GraphTypeID'}      = 5; # line2

            # Create the rrd_graph_item record.
            &createGraphItems($sth2,$graphItem);

        } elsif ($key =~ /PercentFreeVirtMem/) {

            # Set the remaining fields for the graph item.
            $$graphItem{'ColorID'}          = 60;
            $$graphItem{'DSID'}             = $$newDSID{$key};
            $$graphItem{'TextFormat'}       = "Virtual Memory Utilization";
            $$graphItem{'Sequence'}         = 9;
            $$graphItem{'GraphID'}          = $graphIDPercent;
            $$graphItem{'GraphTypeID'}      = 5; # line2

            # Create the rrd_graph_item record.
            &createGraphItems($sth2,$graphItem);

        }
    }
}

#===============================================================================
# graphTypeDisk($dbh,$sth,$sth2,$graph,$graphItem,$hoh_if,$hostname,$newDSID)
#
# Description - Create
#
# Inputs:
#  1.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#=============================================================================
sub graphTypeDisk
{

    my ($dbh,$sth,$sth2,$graph,$graphItem,$hoh_disk,$hostname,$newDSID) = @_;

    my $count = 0;
    my $graphIDPercent;
    my $sequence = 1;

    my @colors          = (96,95,92,90,86,12,22);

    # Create the disk graphs. There will be one PercentFreea graph for all disks
    #  and a BytesFree graph for each disk.
    foreach my $key1 (sort keys %$hoh_disk) {

        foreach my $key2 (sort keys %{$$hoh_disk{$key1}}) {

            if ($key2 =~ /PercentFree/) {

                # On the first pass through, create the PercentFree rrd_graph record.
                if ($count == 0) {

                    # Create the rrd_graph record.
                    $$graph{'Title'}                = $hostname." Disk Utilization";
                    $$graph{'VerticalLabel'}        = "Percent Free";
                    $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
                        or die "\n\nUnable to execute statement handle\n$DBI::errstr";

                    # Save the rrd_graph record ID.
                    $graphIDPercent                 = $dbh->{mysql_insertid};

                    # Save the rrd_graph record ID for use in creating the graph hierarchy.
                    $rrdGraphID{'DiskPercent'}      = $graphIDPercent;

                }

                # Complete the field values for the rrd_graph_item record.
                $$graphItem{'DSID'}                 = $$newDSID{$key2."-".$key1};
                $$graphItem{'ColorID'}              = pop @colors;
                $$graphItem{'TextFormat'}           = "Disk Utilization - Drive ".$$hoh_disk{$key1}{'Descr'};
                $$graphItem{'Sequence'}             = $sequence;
                $$graphItem{'GraphID'}              = $graphIDPercent;
                $$graphItem{'GraphTypeID'}          = 5; # line2

                # Create the rrd_graph_item records.
                &createGraphItems($sth2,$graphItem);

                # Increment the counter(s).
                $count++;;
                $sequence = $sequence + 4;

            } elsif ($key2 =~ /BytesFree/) {

                # Create the rrd_graph record.
                $$graph{'Title'}                    = $hostname." Free Space - Drive ".$$hoh_disk{$key1}{'Descr'};
                $$graph{'VerticalLabel'}            = "Bytes Free";
                $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
                    or die "\n\nUnable to execute statement handle\n$DBI::errstr";

                # Save the rrd_graph record ID for use in creating the graph hierarchy.
                $rrdGraphID{"DiskBytes-".$key1}     = $dbh->{mysql_insertid};

                # Set the rrd_graph_item fields unique to this graph item.
                $$graphItem{'DSID'}                 = $$newDSID{$key2."-".$key1};
                $$graphItem{'ColorID'}              = 22;
                $$graphItem{'TextFormat'}           = "Free Disk Space";
                $$graphItem{'Sequence'}             = 1;
                $$graphItem{'GraphID'}              = $rrdGraphID{"DiskBytes-".$key1};
                $$graphItem{'GraphTypeID'}          = 7; # area


                # Create the rrd_graph_item record.
                &createGraphItems($sth2,$graphItem);

            }
        }
    }
}

#===============================================================================
# graphTypeIf($dbh,$sth,$sth2,$graph,$graphItem,$hoh_if,$hostname,$newDSID)
#
# Description - Create
#
# Inputs:
#  1.
#  2.
#  3.
#  4.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#===============================================================================
sub graphTypeIf
{

    my ($dbh,$sth,$sth2,$graph,$graphItem,$hoh_if,$hostname,$newDSID) = @_;

    # Create the interface graphs. For each interface there will be one for network
    #  traffic (Bytes In/Out) and one for errors and discards
    foreach my $key1 (sort keys %$hoh_if) {

        my $countTraffic    = 0;
        my $graphIDTraffic;

        my $countError      = 0;
        my $graphIDError;

        foreach my $key2 (sort keys %{$$hoh_if{$key1}}) {

            # By matching "Octets" and "Discards|Errors" we eliminate non-interesting
            #  mib values like "ifDescr".
            if ($key2 =~ /Octets/) {

                # On the first pass through, create the network traffic rrd_graph record.
                if ($countTraffic == 0) {

                    # Create the rrd_graph record. (Limit ifDescr to 40 characters)
                    if ($key1 eq "if") {
                        $$graph{'Title'} = $hostname." Interface Traffic";
                    } else {
                        $$graph{'Title'} = $hostname." Interface Traffic - ".substr($$hoh_if{$key1}{'ifDescr'},0,40);;
                    }
                    #$$graph{'Title'}                = $hostname." Interface Traffic";
                    $$graph{'VerticalLabel'}        = "Bytes In/Out";
                    $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
                        or die "\n\nUnable to execute statement handle\n$DBI::errstr";

                    # Save the rrd_graph record ID.
                    $graphIDTraffic                 = $dbh->{mysql_insertid};

                    # Save the rrd_graph record ID for use in creating the graph hierarchy.
                    $rrdGraphID{"IfTraffic-".$key1} = $graphIDTraffic;

                }

                if ($key2 =~ /In/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 22;
                    $$graphItem{'TextFormat'}       = "Bytes In";
                    $$graphItem{'Sequence'}         = 1;
                    $$graphItem{'GraphTypeID'}      = 7; # area

                } elsif ($key2 =~ /Out/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 10;
                    $$graphItem{'TextFormat'}       = "Bytes Out";
                    $$graphItem{'Sequence'}         = 5;
                    $$graphItem{'GraphTypeID'}      = 5; # line2

                }

                # Set rrd_graph_item fields common to all network traffic graphs.
                if ($key1 eq "if") {

                    # This graph is for a switch port. Don't concatenate key1 & key2
                    $$graphItem{'DSID'} = $$newDSID{$key2};
                } else {

                    # This graph is for a server. Merge key1 & key2.
                    $$graphItem{'DSID'} = $$newDSID{$key2."-".$key1};
                }
                $$graphItem{'GraphID'}  = $graphIDTraffic;

                # Create the rrd_graph_item records.
                &createGraphItems($sth2,$graphItem);

                # Increment the counter(s).
                $countTraffic++;

            } elsif ($key2 =~ /Errors|Discards/) {

                # On the first pass through, create the network traffic rrd_graph record.
                if ($countError == 0) {

                    # Create the rrd_graph record. (Limit ifDescr to 40 characters)
                    if ($key1 eq "if") {
                        $$graph{'Title'} = $hostname." Interface Discards/Errors";
                    } else {
                        $$graph{'Title'} = $hostname." Interface Discards/Errors - ".substr($$hoh_if{$key1}{'ifDescr'},0,40);;
                    }
                    #$$graph{'Title'}                = $hostname." Interface Discards/Errors";
                    $$graph{'VerticalLabel'}        = "Bytes In/Out";
                    $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
                        or die "\n\nUnable to execute statement handle\n$DBI::errstr";

                    # Save the rrd_graph record ID.
                    $graphIDError                   = $dbh->{mysql_insertid};

                    # Save the rrd_graph record ID for use in creating the graph hierarchy.
                    $rrdGraphID{"IfErrors-".$key1}  = $graphIDError;
                }

                if ($key2 =~ /ifInDiscards/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 25;
                    $$graphItem{'TextFormat'}       = "Discards In";
                    $$graphItem{'Sequence'}         = 1;
                    $$graphItem{'GraphTypeID'}      = 7; # area

                } elsif ($key2 =~ /ifInErrors/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 35;
                    $$graphItem{'TextFormat'}       = "Errors In";
                    $$graphItem{'Sequence'}         = 5;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /ifOutDiscards/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 9;
                    $$graphItem{'TextFormat'}       = "Discards Out";
                    $$graphItem{'Sequence'}         = 9;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /ifOutErrors/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 11;
                    $$graphItem{'TextFormat'}       = "Errors Out";
                    $$graphItem{'Sequence'}         = 13;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                }

                # Set rrd_graph_item fields common to all network error graphs.
                if ($key1 eq "if") {
                    $$graphItem{'DSID'} = $$newDSID{$key2};
                } else {
                    $$graphItem{'DSID'} = $$newDSID{$key2."-".$key1};
                }
                $$graphItem{'GraphID'}  = $graphIDError;

                # Create the rrd_graph_item records.
                &createGraphItems($sth2,$graphItem);

                # Increment the counter(s).
                $countError++;

            }
        }
    }
}

#===============================================================================
# graphTypePrn($dbh,$sth,$sth2,$graph,$graphItem,$hoh_if,$hostname,$newDSID)
#
# Description - Create
#
# Inputs:
#  1.
#  2.
#  3.
#  4.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#===============================================================================
sub graphTypePrn
{

    my ($dbh,$sth,$sth2,$graph,$graphItem,$hoh_prn,$hostname,$newDSID) = @_;

    # Create the Printer page count graph. There will be one for each printer.
    foreach my $key (keys %{$$hoh_prn{'Printer'}}) {

        # Create the rrd_graph record.
        $$graph{'Title'}                = $hostname." Page Count";
        $$graph{'VerticalLabel'}        = "Pages Printed";
        $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";

        # Set the rrd_graph_item default field values.
        $$graphItem{'GraphID'}          = $dbh->{mysql_insertid};
        $$graphItem{'GraphTypeID'}      = 5; # line2

        # Save the rrd_graph record ID for use in creating the graph hierarchy.
        $rrdGraphID{'Printer'}          = $$graphItem{'GraphID'};

        # Complete the settings for this graph item.
        $$graphItem{'DSID'}             = $$newDSID{$key};
        $$graphItem{'ColorID'}          = 21; #0000FF Blue
        $$graphItem{'TextFormat'}       = "Printer Page Count";
        $$graphItem{'Sequence'}         = 1;

        # Create the rrd_graph_item records.
        &createGraphItems($sth2,$graphItem);

    }
}

#===============================================================================
# graphTypeRmon($dbh,$sth,$sth2,$graph,$graphItem,$hoh_if,$hostname,$newDSID)
#
# Description - Create
#
# Inputs:
#  1.
#  2.
#  3.
#  4.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#===============================================================================
sub graphTypeRmon
{

    my ($dbh,$sth,$sth2,$graph,$graphItem,$hoh_rmon,$hostname,$newDSID) = @_;

    # For testing purposes only.
    #dref($hoh_if,"hoh");

    # Create the interface graphs. For each interface there will be one for network
    #  traffic (Bytes In/Out) and one for errors and discards
    foreach my $key1 (sort keys %$hoh_rmon) {

        my $countPktSize    = 0;
        my $graphIDPktSize;

        my $countError      = 0;
        my $graphIDError;

        foreach my $key2 (sort keys %{$$hoh_rmon{$key1}}) {

            # Create the packet size distribution graph.
            if ($key2 =~ /(64|to)/) {

                # On the first pass through, create the rrd_graph record.
                if ($countPktSize == 0) {

                    # Create the rrd_graph record.
                    $$graph{'Title'}                = $hostname." RMON Packet Size Distribution";
                    $$graph{'VerticalLabel'}        = "Packets";
                    $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
                        or die "\n\nUnable to execute statement handle\n$DBI::errstr";

                    # Save the rrd_graph record ID.
                    $graphIDPktSize                 = $dbh->{mysql_insertid};

                    # Save the rrd_graph record ID for use in creating the graph hierarchy.
                    $rrdGraphID{'PktSize'}          = $graphIDPktSize;

                }

                if ($key2 =~ /64/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 21;
                    $$graphItem{'TextFormat'}       = "64 Byte Packets";
                    $$graphItem{'Sequence'}         = 1;
                    $$graphItem{'GraphTypeID'}      = 7; # area

                } elsif ($key2 =~ /65to127/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 7;
                    $$graphItem{'TextFormat'}       = "65-127 Byte Packets";
                    $$graphItem{'Sequence'}         = 5;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /128to255/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 80;
                    $$graphItem{'TextFormat'}       = "128-255 Byte Packets";
                    $$graphItem{'Sequence'}         = 9;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /256to511/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 64;
                    $$graphItem{'TextFormat'}       = "256-511 Byte Packets";
                    $$graphItem{'Sequence'}         = 13;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /512to1023/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 58;
                    $$graphItem{'TextFormat'}       = "512-1023 Byte Packets";
                    $$graphItem{'Sequence'}         = 17;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /1024to1518/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 57;
                    $$graphItem{'TextFormat'}       = "1024-1518 Byte Packets";
                    $$graphItem{'Sequence'}         = 21;
                    $$graphItem{'GraphTypeID'}      = 8; # stack
                }

                # Set rrd_graph_item fields common to all network traffic graphs.
                $$graphItem{'DSID'}     = $$newDSID{$key2};
                $$graphItem{'GraphID'}  = $graphIDPktSize;

                # Create the rrd_graph_item records.
                &createGraphItems($sth2,$graphItem);

                # Increment the counter(s).
                $countPktSize++;

            # Create the error graph.
            } elsif ($key2 =~ /(CRC|Undersize|Oversize|Fragments|Jabbers)/) {

                # On the first pass through, create the rrd_graph record.
                if ($countError == 0) {

                    # Create the rrd_graph record.
                    $$graph{'Title'}                = $hostname." RMON Ethernet Errors";
                    $$graph{'VerticalLabel'}        = "Errors";
                    $sth->execute ($$graph{'Title'},$$graph{'VerticalLabel'})
                        or die "\n\nUnable to execute statement handle\n$DBI::errstr";

                    # Save the rrd_graph record ID.
                    $graphIDError                   = $dbh->{mysql_insertid};

                    # Save the rrd_graph record ID for use in creating the graph hierarchy.
                    $rrdGraphID{'Errors'}           = $graphIDError;
                }

                if ($key2 =~ /CRC/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 25;
                    $$graphItem{'TextFormat'}       = "CRC Errors";
                    $$graphItem{'Sequence'}         = 1;
                    $$graphItem{'GraphTypeID'}      = 7; # area

                } elsif ($key2 =~ /Undersize/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 35;
                    $$graphItem{'TextFormat'}       = "Undersized Packets";
                    $$graphItem{'Sequence'}         = 5;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /Oversize/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 9;
                    $$graphItem{'TextFormat'}       = "Oversized Packets";
                    $$graphItem{'Sequence'}         = 9;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /Fragments/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 11;
                    $$graphItem{'TextFormat'}       = "Fragments";
                    $$graphItem{'Sequence'}         = 13;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                } elsif ($key2 =~ /Jabbers/) {

                    # Set the rrd_graph_item fields unique to this graph item.
                    $$graphItem{'ColorID'}          = 58;
                    $$graphItem{'TextFormat'}       = "Jabbers";
                    $$graphItem{'Sequence'}         = 17;
                    $$graphItem{'GraphTypeID'}      = 8; # stack

                }

                # Set rrd_graph_item fields common to all network traffic graphs.
                $$graphItem{'DSID'}     = $$newDSID{$key2};
                $$graphItem{'GraphID'}  = $graphIDError;

                # Create the rrd_graph_item records.
                &createGraphItems($sth2,$graphItem);

                # Increment the counter(s).
                $countError++;

            }
        }
    }
}

#=============================================================================
# createGraphItems($sth2,$graphItem)
#
# Description - Create the rrd_graph_item records.
#
# Inputs:
#  1.$sth2
#  1.$graphItem
#  1.$sequence
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#=============================================================================
sub createGraphItems
{

    my ($sth2,$graphItem)   = @_;

    $$graphItem{'ConsolidationFunction'}    = 1;            # 1 = avg; 3 = max; 4 = last
    $$graphItem{'HardReturn'}               = "";           #"on" or ""(off)

    # Create the first graph.
    $sth2->execute ($$graphItem{'DSID'},$$graphItem{'ColorID'},$$graphItem{'TextFormat'},
        $$graphItem{'Sequence'},$$graphItem{'GraphID'},$$graphItem{'GraphTypeID'},
        $$graphItem{'ConsolidationFunction'},$$graphItem{'HardReturn'},$$graphItem{'CDEFID'},
        $$graphItem{'SequenceParent'},$$graphItem{'Parent'},$$graphItem{'GprintOpts'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";

    # Modify the graph item hash for subsequent gprint items
    $$graphItem{'ColorID'}                  = 0;           # not used for gprint graphs
    $$graphItem{'GraphTypeID'}              = 9;           # gprint

    # Add "Current" GPRINT graph item
    $$graphItem{'TextFormat'}               = "Current:";
    $$graphItem{'Sequence'}++;
    $$graphItem{'ConsolidationFunction'}    = 4;           # 1 = avg; 3 = max; 4 = last

    $sth2->execute ($$graphItem{'DSID'},$$graphItem{'ColorID'},$$graphItem{'TextFormat'},
        $$graphItem{'Sequence'},$$graphItem{'GraphID'},$$graphItem{'GraphTypeID'},
        $$graphItem{'ConsolidationFunction'},$$graphItem{'HardReturn'},$$graphItem{'CDEFID'},
        $$graphItem{'SequenceParent'},$$graphItem{'Parent'},$$graphItem{'GprintOpts'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";

    # Add "Average" GPRINT graph item
    $$graphItem{'TextFormat'}               = "Average:";
    $$graphItem{'Sequence'}++;
    $$graphItem{'ConsolidationFunction'}    = 1;           # 1 = avg; 3 = max; 4 = last

    $sth2->execute ($$graphItem{'DSID'},$$graphItem{'ColorID'},$$graphItem{'TextFormat'},
        $$graphItem{'Sequence'},$$graphItem{'GraphID'},$$graphItem{'GraphTypeID'},
        $$graphItem{'ConsolidationFunction'},$$graphItem{'HardReturn'},$$graphItem{'CDEFID'},
        $$graphItem{'SequenceParent'},$$graphItem{'Parent'},$$graphItem{'GprintOpts'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";

    # Add "Maximum" GPRINT graph item
    $$graphItem{'TextFormat'}               = "Maximum:";
    $$graphItem{'Sequence'}++;
    $$graphItem{'ConsolidationFunction'}    = 3;            # 1 = avg; 3 = max; 4 = last
    $$graphItem{'HardReturn'}               = "on";         #"on" or ""(off)

    $sth2->execute ($$graphItem{'DSID'},$$graphItem{'ColorID'},$$graphItem{'TextFormat'},
        $$graphItem{'Sequence'},$$graphItem{'GraphID'},$$graphItem{'GraphTypeID'},
        $$graphItem{'ConsolidationFunction'},$$graphItem{'HardReturn'},$$graphItem{'CDEFID'},
        $$graphItem{'SequenceParent'},$$graphItem{'Parent'},$$graphItem{'GprintOpts'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";

}

#=============================================================================
# createGraphHierarchy($dbh,$rrdGraphID,$heading,$hostname)
#
# Description - Create
#
# Inputs:
#  1.$dbh           - Database handle to cacti database.
#  2.$rrdGraphID    - A reference to a hash containing the rrd_ds field values.
#  3.$heading       - A reference to a hash containing the rrd_ds field values.
#  4.$hostname      - The name of the device. This will be the name under which the device
#                       will be created. The default is the sysName for SNMP devices and
#                       the Windows machine name for WMI-polled devices.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#=============================================================================
sub createGraphHierarchy
{

    my ($dbh,$rrdGraphID,$heading,$hostname) = @_;

    my ($query,$sth,$sth2,$result);

    # Create a hash of the rrd_graph table default field values.
    my %treeItem;
    #$treeItem{'TreeID'}                                # This will be defined later.
    $treeItem{'GraphID'}         = 0;                   # '0' for headings.
    $treeItem{'RRAID'}           = 1;                   # Daily (5-minute average)
    #$treeItem{'Title'}                                 # This will be defined later.
    $treeItem{'Type'}            = "Heading";           # 'Heading' or 'Graph'
    $treeItem{'Parent'}          = 0;                   # Root of tree
    #$treeItem{'Sequence'}                              # This will be defined later.

    # Use the default hierarchy 'Auto-created', but first check to see
    #  if it already exists.
    $query = "SELECT ID FROM graph_hierarchy WHERE Name = 'Auto-created'";
    my $result = $dbh->selectrow_array ($query);
    if (defined($result)) {

        # It already exists. Now save the ID.
        $treeItem{'TreeID'} = $result;

    } else {

        # It doesn't exist. Create the default hierarchy.
        $query = "INSERT INTO graph_hierarchy (Name) VALUES ('Auto-created')";
        $dbh->do ($query)
                or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
        $treeItem{'TreeID'} = $dbh->{mysql_insertid};

    }

    # Test to see if the heading already exists; reuse it if it does.
    $treeItem{'Title'} = $heading;
    $query = "SELECT ID FROM graph_hierarchy_items WHERE Title = '$treeItem{'Title'}'";
    $result = $dbh->selectrow_array ($query);

    if (defined($result)) {

        # The heading already exists. Use the existing ID.
        $treeItem{'Parent'} = $result;

    } else {

        # Otherwise, create a new heading.
        # Start by finding the maximum sequence # already in use at
        #  the root of the tree.
        $query = "SELECT MAX(Sequence) FROM graph_hierarchy_items
                    WHERE TreeID = $treeItem{'TreeID'} AND Parent = 0";
        my $sequence = $dbh->selectrow_array ($query);
        if (defined($sequence)) {
            $treeItem{'Sequence'} = $sequence++;
        } else {
            $treeItem{'Sequence'} = 1;
        }

        # Next, create a graph_hierarchy_item heading record.
        $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,Title,
                    Type,Parent,Sequence)
                    VALUES ($treeItem{'TreeID'},$treeItem{'GraphID'},$treeItem{'RRAID'},
                    '$treeItem{'Title'}','$treeItem{'Type'}',$treeItem{'Parent'},
                    $treeItem{'Sequence'})";
        $dbh->do ($query)
            or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
        $treeItem{'Parent'} = $dbh->{mysql_insertid};

    }

    # Next create heading for this host below $heading
    # Start by finding the maximum sequence # already in use
    $query = "SELECT MAX(Sequence) FROM graph_hierarchy_items
                WHERE TreeID = $treeItem{'TreeID'} AND Parent = $treeItem{'Parent'}";
    my $sequence = $dbh->selectrow_array ($query);
    if (defined($sequence)) {
        $treeItem{'Sequence'} = $sequence++;
    } else {
        $treeItem{'Sequence'} = 1;
    }

    # Next, create a graph_hierarchy_item heading record.
    $treeItem{'Title'} = $hostname;
    $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,Title,
                Type,Parent,Sequence)
                VALUES ($treeItem{'TreeID'},$treeItem{'GraphID'},$treeItem{'RRAID'},
                '$treeItem{'Title'}','$treeItem{'Type'}',$treeItem{'Parent'},
                $treeItem{'Sequence'})";
    $dbh->do ($query)
        or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
    $treeItem{'Parent'} = $dbh->{mysql_insertid};


    # Modify the treeItem hash for the graphs that will be inserted
    #  into the hierarchy tree.
    $treeItem{'Type'}            = "Graph";             # 'Heading' or 'Graph'
    $treeItem{'Sequence'}        = 1;                   # Starting sequence number

    # Finally, insert the actual graphs in the hierarchy.
    $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,
                Type,Parent,Sequence)
                VALUES ($treeItem{'TreeID'},?,$treeItem{'RRAID'},
                '$treeItem{'Type'}',$treeItem{'Parent'},?)";
    $sth = $dbh->prepare ($query)
        or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";
    foreach my $value (sort values %$rrdGraphID) {
        $treeItem{'GraphID'} = $value;
        $sth->execute ($treeItem{'GraphID'},$treeItem{'Sequence'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
        $treeItem{'Sequence'}++;
    }

}

#=============================================================================
# createGraphHierarchyTree($dbh,$rrdGraphID,$hierarchy)
#
# Description - Create
#
# Inputs:
#  1.$dbh           - Database handle to cacti database.
#  2.$rrdGraphID    - A reference to a hash containing the rrd_ds field values.
#  3.$hierarchy     - A reference to a hash containing graph hierarchy settings.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#=============================================================================
sub createGraphHierarchyTree
{

    my ($dbh,$rrdGraphID,$hierarchy) = @_;

    my ($query,$sth,$sth2,$result);

    # Create a hash of the rrd_graph table default field values.
    my %treeItem;
    #$treeItem{'TreeID'}                            # This will be defined later.
    $treeItem{'GraphID'}    = 0;                    # '0' for headings.
    $treeItem{'RRAID'}      = 1;                    # Daily (5-minute average)
    #$treeItem{'Title'}                             # This will be defined later.
    $treeItem{'Type'}       = "Heading";            # 'Heading' or 'Graph'
    $treeItem{'Parent'}     = 0;                    # Root of tree
    #$treeItem{'Sequence'}                          # This will be defined later.

    # Create the root of the tree.
    $query = "SELECT ID FROM graph_hierarchy WHERE Name = '$$hierarchy{'root'}'";
    $result = $dbh->selectrow_array ($query);
    if (defined($result)) {

        # The root already exists. Now save the ID.
        $treeItem{'TreeID'} = $result;

    } else {

        # It doesn't exist. Create the root.
        $query = "INSERT INTO graph_hierarchy (Name) VALUES ('$$hierarchy{'root'}')";
        $dbh->do ($query)
                or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
        $treeItem{'TreeID'} = $dbh->{mysql_insertid};

    }


    # Create the heading.
    $treeItem{'Title'} = $$hierarchy{'heading'};
    $query = "SELECT ID FROM graph_hierarchy_items WHERE Title = '$treeItem{'Title'}'";
    $result = $dbh->selectrow_array ($query);
    if (defined($result)) {

        # The heading already exists. Use the existing ID.
        $treeItem{'Parent'} = $result;

    } else {

        # It doesn't exist. Create the heading.
        # Start by finding the maximum sequence # already in use
        $query = "SELECT MAX(Sequence) FROM graph_hierarchy_items
                    WHERE TreeID = $treeItem{'TreeID'} AND Parent = 0";
        my $sequence = $dbh->selectrow_array ($query);
        if (defined($sequence)) {
            $treeItem{'Sequence'} = $sequence++;
        } else {
            $treeItem{'Sequence'} = 1;
        }

        # Next, create a graph_hierarchy_item heading record.
        $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,Title,
                    Type,Parent,Sequence)
                    VALUES ($treeItem{'TreeID'},$treeItem{'GraphID'},$treeItem{'RRAID'},
                    '$treeItem{'Title'}','$treeItem{'Type'}',$treeItem{'Parent'},
                    $treeItem{'Sequence'})";
        $dbh->do ($query)
            or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
        $treeItem{'Parent'} = $dbh->{mysql_insertid};

    }

    # In some cases (e.g., switches) a hostname heading will need to be created too.
    #  This is indicated by the presence of the hostname entry in the %hierarchy hash.
    if (defined($$hierarchy{'hostname'})) {
        $treeItem{'Title'} = $$hierarchy{'hostname'};
        $query = "SELECT ID FROM graph_hierarchy_items WHERE Title = '$treeItem{'Title'}'";
        $result = $dbh->selectrow_array ($query);
        if (defined($result)) {

            # The hostname already exists. Use the existing ID.
            $treeItem{'Parent'} = $result;

        } else {

            # It doesn't exist. Create the heading.
            # Start by finding the maximum sequence # already in use
            $query = "SELECT MAX(Sequence) FROM graph_hierarchy_items
                        WHERE TreeID = $treeItem{'TreeID'} AND Parent = $treeItem{'Parent'}";
            my $sequence = $dbh->selectrow_array ($query);
            if (defined($sequence)) {
                $treeItem{'Sequence'} = $sequence++;
            } else {
                $treeItem{'Sequence'} = 1;
            }

            # Next, create a graph_hierarchy_item hostname record.
            $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,Title,
                        Type,Parent,Sequence)
                        VALUES ($treeItem{'TreeID'},$treeItem{'GraphID'},$treeItem{'RRAID'},
                        '$treeItem{'Title'}','$treeItem{'Type'}',$treeItem{'Parent'},
                        $treeItem{'Sequence'})";
            $dbh->do ($query)
                or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
            $treeItem{'Parent'} = $dbh->{mysql_insertid};

        }
    }

    # Create the device heading.
    $treeItem{'Title'} = $$hierarchy{'device'};
    $query = "SELECT ID FROM graph_hierarchy_items WHERE Title = '$treeItem{'Title'}'";
    $result = $dbh->selectrow_array ($query);
    if (defined($result)) {

        # WARNING!!! The device already exists. Use the existing ID.
        $treeItem{'Parent'} = $result;

    } else {

        # It doesn't exist. Create the heading.
        # Start by finding the maximum sequence # already in use
        $query = "SELECT MAX(Sequence) FROM graph_hierarchy_items
                    WHERE TreeID = $treeItem{'TreeID'} AND Parent = $treeItem{'Parent'}";
        my $sequence = $dbh->selectrow_array ($query);
        if (defined($sequence)) {
            $treeItem{'Sequence'} = $sequence++;
        } else {
            $treeItem{'Sequence'} = 1;
        }

        # Next, create a graph_hierarchy_item heading record.
        $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,Title,
                    Type,Parent,Sequence)
                    VALUES ($treeItem{'TreeID'},$treeItem{'GraphID'},$treeItem{'RRAID'},
                    '$treeItem{'Title'}','$treeItem{'Type'}',$treeItem{'Parent'},
                    $treeItem{'Sequence'})";
        $dbh->do ($query)
            or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
        $treeItem{'Parent'} = $dbh->{mysql_insertid};
    }


    # Modify the treeItem hash for the graphs that will be inserted
    #  into the hierarchy tree.
    $treeItem{'Type'}            = "Graph";             # 'Heading' or 'Graph'
    $treeItem{'Sequence'}        = 1;                   # Starting sequence number

    # Finally, insert the actual graphs in the hierarchy.
    $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,
                Type,Parent,Sequence)
                VALUES ($treeItem{'TreeID'},?,$treeItem{'RRAID'},
                '$treeItem{'Type'}',$treeItem{'Parent'},?)";
    $sth = $dbh->prepare ($query)
        or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";
    foreach my $value (sort values %$rrdGraphID) {
        $treeItem{'GraphID'} = $value;
        $sth->execute ($treeItem{'GraphID'},$treeItem{'Sequence'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
        $treeItem{'Sequence'}++;
    }

}

#=============================================================================
# createGraphHierarchyItem($dbh,$rrdGraphID,$hierarchy)
#
# Description - Create
#
# Inputs:
# Inputs:
#  1.$dbh           - Database handle to cacti database.
#  2.$rrdGraphID    - A reference to a hash containing the rrd_ds field values.
#  3.$hierarchy     - A reference to a hash containing graph hierarchy settings.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#=============================================================================
sub createGraphHierarchyItem
{

    my ($dbh,$rrdGraphID,$hierarchy) = @_;

    my ($query,$sth,$sth2,$result);

    # Create a hash of the rrd_graph table default field values.
    my %treeItem;
    #$treeItem{'TreeID'}                                # This will be defined later.
    $treeItem{'GraphID'}         = 0;                   # '0' for headings.
    $treeItem{'RRAID'}           = 1;                   # Daily (5-minute average)
    #$treeItem{'Title'}                                 # This will be defined later.
    $treeItem{'Type'}            = "Heading";           # 'Heading' or 'Graph'
    $treeItem{'Parent'}          = 0;                   # Root of tree
    #$treeItem{'Sequence'}                              # This will be defined later.

    # Test to see if the heading already exists; reuse it if it does.
    $treeItem{'Title'} = $$hierarchy{'heading'};
    $query = "SELECT ID FROM graph_hierarchy_items WHERE Title = '$treeItem{'Title'}'";
    $result = $dbh->selectrow_array ($query);

    if (defined($result)) {

        # The heading already exists. Use the existing ID.
        $treeItem{'Parent'} = $result;

    } else {

        # Otherwise, create a new heading.
        # Start by finding the maximum sequence # already in use at
        #  the root of the tree.
        $query = "SELECT MAX(Sequence) FROM graph_hierarchy_items
                    WHERE TreeID = $treeItem{'TreeID'} AND Parent = 0";
        my $sequence = $dbh->selectrow_array ($query);
        if (defined($sequence)) {
            $treeItem{'Sequence'} = $sequence++;
        } else {
            $treeItem{'Sequence'} = 1;
        }

        # Next, create a graph_hierarchy_item heading record.
        $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,Title,
                    Type,Parent,Sequence)
                    VALUES ($treeItem{'TreeID'},$treeItem{'GraphID'},$treeItem{'RRAID'},
                    '$treeItem{'Title'}','$treeItem{'Type'}',$treeItem{'Parent'},
                    $treeItem{'Sequence'})";
        $dbh->do ($query)
            or die "\n\nUnable to execute query:\n$query\n$DBI::errstr";
        $treeItem{'Parent'} = $dbh->{mysql_insertid};

    }

    # Modify the treeItem hash for the graphs that will be inserted
    #  into the hierarchy tree.
    $treeItem{'Type'}            = "Graph";             # 'Heading' or 'Graph'
    $treeItem{'Sequence'}        = 1;                   # Starting sequence number

    # Finally, insert the actual graphs in the hierarchy.
    $query = "INSERT INTO graph_hierarchy_items (TreeID,GraphID,RRAID,
                Type,Parent,Sequence)
                VALUES ($treeItem{'TreeID'},?,$treeItem{'RRAID'},
                '$treeItem{'Type'}',$treeItem{'Parent'},?)";
    $sth = $dbh->prepare ($query)
        or die "\n\nUnable to prepare query:\n$query\n$DBI::errstr";
    foreach my $value (sort values %$rrdGraphID) {
        $treeItem{'GraphID'} = $value;
        $sth->execute ($treeItem{'GraphID'},$treeItem{'Sequence'})
            or die "\n\nUnable to execute statement handle\n$DBI::errstr";
        $treeItem{'Sequence'}++;
    }

}

#=============================================================================
