#===============================================================================
# Program:      subs_misc.pl
#
# Description:  Library of miscellaneous subroutines.
#
#               This script was developed under ActiveState Perl version 5.6.1.
#
# Requirements: This ActiveState PERL must be installed on the system where the
#               script is executed.
#
#               The following additional Perl modules must also be installed:
#                   None
#
#               The following files must exist in the same directory as this
#               script:
#                   None
#
#               Known limitations:
#                   None
#
#
# Date       Author/Email                Description
# ---------- --------------------------- ---------------------------------------
# 03/14/2003 Jim Lancaster               Initial version 0.1
#            jlancaster@mbdworld.com
#
# 03/17/2003 Jim Lancaster               Version 0.1a
#                                        Fixed issues with snmpTable
#                                        Stripped out cacti subs into separate
#                                         script file.
#                                        Stripped out SNMP subs and combined
#                                         with mibs.pl in new script file.
#
# 03/17/2003 Jim Lancaster               Version 0.1b
#                                        Modified usage() slightly.
#
#
# Known issues:
#    1. None
#
# Future plans:
#    1. None
#
#===============================================================================
use strict;
return 1;

#===============================================================================
#                       Subroutines
#===============================================================================
# usage($usageText)
#
# Description - Print usage and exit.
#
# Inputs:
#  1. $usageText - Usage instructions.
#
# Processing:
#  1.
#
# Outputs:
#  1.
#
#===============================================================================
sub usage
{
        my ($usageText) = @_;
        if($usageText ne "") {
            die "$usageText\n";
        }

        die "No usage information available.\n";
}

#===============================================================================
# dref($ref,$type)
#
# Description - De-reference an array, hash, or hash of hashes. Used mostly
#                for testing purposes.
#
# Inputs:
#  1. $ref - Reference to an array, hash, or hash of hashes.
#  2. $type - Either "array","hash", or "hoh".
#
# Processing:
#  1.
#
# Outputs:
#  1. Print out contents.
#
#===============================================================================
sub dref
{
    my $ref = shift;
    my $type = shift;

    if ($type eq "array") {
        foreach my $value (@$ref) {
            print "$value\n";
        }
    } elsif ($type eq "hash") {
        foreach my $key1 (sort keys %$ref) {
            print "$key1 => $$ref{$key1}\n";
        }
    } elsif ($type eq "hoh") {
        foreach my $key1 (sort keys %$ref) {
            foreach my $key2 (sort keys %{$$ref{$key1}}) {
                print "$key1 => $key2 => $$ref{$key1}{$key2}\n";
            }
        }
    } else {
        warn ("Unknown reference type.\n");
    }

}

#===============================================================================
# readCfg($filename)
#
# Description - Read configuration settings from a cfg file.
#
# Inputs:
#  1. $filename - config filename.
#
# Processing:
#  1. Open config file, read data, store in hash.
#
# Outputs:
#  1. Hash containing the setup information
#
#===============================================================================
sub readCfg
{

    my ($filename) = shift;
    my (@array,%config);

    # Open the config file. If file not found, error out.
    open(CFGFILE,"<$filename")
        or open(CFGFILE,"<scripts/$filename")
            or die "\n\nUnable to open configuration file $filename!";

    # Strip off comments and blank lines, leaving an array of config data.
    while (my $line = <CFGFILE>) {

        # Skip comment lines.
        next if (substr ($line,0,1) eq "#");

        # Skip blank lines.
        next if ($line eq "");

        # Remove carriage return/line feed from EOL.
        chop($line);

        # Stuff the remaining data into a hash.
        @array = split(/=/,$line);
        if ($#array == 1) {
            %config = (%config,$array[0],$array[1]);
        }
    }

    close(CFGFILE);

    return (%config);
}

#=============================================================================
